/**
* serial_controller.ido
* 
* ===========
* DESCRIPTION
* ===========
*
* This is an Arduino sketch to control the ArcBotics Sparki robot via Bluetooth.
* There are several such programs availble.  This differs from the others I've
* found in that the program is set up to be a bit more "programmatically close"
* to higher level languages.  This allows students to see parameterization of
* the incoming Bluetooth stream.  The protocol is as follows:
* Incoming data packets are fixed at 13 characters (MAX_BUFFER + 1) wide and
* they have the format:
*     CDDDDDDDDDDDD
* where "C" represents a single character code for the function and "D"
* represents additional data for the given function.
* 
* C MUST be one of the codes outlined in the switch statement in this code:
*    FBLRSCOPGHV+\|/lbr
* D MUST fall into one of three possibilities:
*     0-9: representing a numeric value or the continuation thereof of the
*          current parameter
*     ,:   representing a separation to the next parameter
*     _:   representing no more data
*
* Examples:
* - Move forward until stop (sparki.moveForward()):
*     F____________
* - Move forward 100 (sparki.moveForward(100)):
*     F100_________
* - Move backward 100 (sparki.moveBackward(100)):
*     B100_________
* - Set the LED to R,G,B values 50,100,20 (sparki.RGB(50,100,20)):
*     +50,100,20___
*
* In all cases, Sparki will show the code, parameters, and parameter count on the
* LCD.  This also has a built in "gun" using the IR sender/receiver.  "Shooting"
* IR code 28 at another Sparki running the same code will register a "hit" on the
* robot.  The idea being here that one could set up a game out of this with some
* as an extension.
* 
* At the end of the execution of any given command, Sparki will send back (most of)
* the sensor data via Bluetooth in a variable-width string of the following format:
*     {AX,AY,AZ,MX,MY,MZ,C,P,ILL,IL,IC,IR,IRR,LL,LC,LR}
* 
* This sequence includes the curly braces and is terminated by a newline ("\n")
* The mapping of the above codes is as follows:
*     AX : Accelerometer in X direction
*     AY : Accelerometer in Y direction
*     AZ : Accelerometer in Z direction
*     MX : Magnetometer in X direction
*     MY : Magnetometer in Y direction
*     MZ : Magnetometer in Z direction
*     C  : Compass
*     P  : Ultrasonic distance
*     ILL: Far left edge reflective sensor
*     IL : Left of center reflective sensor
*     IC : Center reflective sensor
*     IR : Right of center reflective sensor
*     IRR: Far right edge reflective sensor
*     LL : Left light sensor
*     LC : Center light sensor
*     LR : Right light sensor
* 
* Please note that the LCD is updated and this sequence is sent out via Bluetooth
* AFTER any blocking commands have COMPLETED.
* 
* As I was writing this code, I ran out of space in the Arduino at several times
* and had to reduce the overall sketch size by taking out several "unnecessary"
* functions.  For example, when mapping parameters, it would be ideal to use the
* String library with the simple toInt() function, but including it adds lots of
* overhead space, so if you look through the code, you may notice things that
* look particularly ugly but serve the purpose of reducing code space.  I have
* attempted to document several of them below.  If you have additional suggestions
* for ways to reduce the compiled code size, please let me know!
* 
* Slight gripe: I've found it frustrating that the Arduino compiler/linker doesn't
* allow you to see the total code output size if/when it's too large, but instead
* just wipes it off disk immediately.  This makes it really difficult to know how
* close you are to bringing a sketch down to allowable size.
*
* =========
* DONATIONS
* =========
* If you like this software, I'm glad!  I'm also presently unemployed and if you'd
* like to make a donation of any size to help me out, I would surely appreciate it!
*
* BitCoin: 1PAn1UMmBzJsKTJKKzAYCMoFmPtPgz4NGR
* PayPal:  shane@eskibars.com
* 
* ==============
* SPECIAL THANKS
* ==============
*
* I would like to thank all of those that have helped make this software and the
* Sparki hardware platform available and open.  In particular:
* 1) Thanks to the Arbotics team for creating a fantastic platform for low-cost
* and exciting experimentation.  I know when I was first learning to write
* software, the lack of "exciting" problems to dive into quickly was incredibly
* frustrating and I was merely lucky to have some great mentors at the time to
* push me through.
* 2) Thanks to the SJ Davis Robotics team for coming up with a great bluetooth
* controller for Sparki.  Though this project was a from-scratch writing and thus
* doesn't share direct code with that project, I did get some of the ideas from
* their excellent application, which at this time is hosted at
* https://github.com/sivads/SJD_BT_Control
* 3) Thanks to Julian Golder for providing many of the ideas for this project
* using his extensive educational background
* 4) Thanks to my family and many friends for their support and putting up with
* my annoying technical mumbo-jumbo as I work out technical difficulties
* 5) Thanks to my alma mater, Georgia Tech, and in particular, Dr. Thomas R Collins,
* who provided several of us the opportunity to be lab TAs in his own exciting
* lab-based course which involved robotics.  There are few things that get
* enterprising students as excited as the ability to teach others, and I was
* one of those lucky enough to do so.  This ultimately got me excited about
* robotics and teaching that and software.
*
* ================
* COPYRIGHT NOTICE
* ================
* 
* The MIT License
* 
* Copyright (c) 2014 Shane Connelly (shane@eskibars.com)
* 
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
* 
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
* 
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
* THE SOFTWARE.
**/

#include <Sparki.h> // include the sparki library

#define MAX_BUFFER 12 // the maximum number of characters for additional parameters in the transmission

int hits = 0;
int servoPosition = SERVO_CENTER;
char currentLightColor = 0;
int paramCount = 0;
boolean newData = true;

void setup()
{
  Serial.begin(9600);
  Serial1.begin(9600);
  sparki.servo(SERVO_CENTER);
  delay(100); // delay to establish the usb connection. using the arduino while-serial command doesn't 
                // let it boot properly when there's no USB conenction.
}

// I know this looks like a really stupid function, but it saves a few precious bytes of space rather than having the constants in a loop below
void writeSeparator()
{
  Serial1.write('_');
}

char serialCode = '0';

void loop()
{
  char additionalParams [MAX_BUFFER];
  int IRCode = sparki.readIR();
 
  if(IRCode != -1)
  {
    newData = true;
    if (IRCode == 28)
    {
      hits++;
    }
  }
  
  if(Serial1.available() && ((char)Serial1.peek()) != '<')
  {
    serialCode = (char)Serial1.read();
    newData = true;
    char additionalCodeCount = 0;
    paramCount = 0;
    delay(10); // wait to get the full packet in before reading any more
    int paramArray[MAX_BUFFER];
    
    char startChar = 0;
    int val = 0;
    boolean sequenceGoing = true;
    for (int i = 0; i < MAX_BUFFER; i++)
    {
      paramArray[i] = 0;
    }
    while(additionalCodeCount < MAX_BUFFER && Serial1.available())
    {
      char res = (char)Serial1.read();
      additionalParams[additionalCodeCount] = res;
      
      if (sequenceGoing && additionalCodeCount > 0 && (res == '_' || res == ','))
      {
        /*
        // the String library takes too much space in the program so have to take it out and do manual processing of the string -> int
        String tNumStr = String(additionalParams).substring(startChar, additionalCodeCount-1);
        int val = tNumStr.toInt();
        */
        paramArray[paramCount] = val;
        paramCount++;
        val = 0;
      }
      else if (sequenceGoing > 0 && res != '_' && res != ',')
      {
        val *= 10;
        val += (res - '0');
      }
      
      additionalCodeCount++;
      if (res == '_')
      {
        sequenceGoing = false;
      }
      
      delay(1);
    }
    if (sequenceGoing)
    {
      paramCount++;
      paramArray[paramCount] = val;
    }
    Serial1.flush();
    
    // we can save a bit of space in the Arduino by doing this here for the few times it comes up
    int param1multiple = paramArray[1] * 10;
    switch(serialCode)
    {
      case 'F':
        if (paramCount > 0)
        {
          int amount = paramArray[0];
          sparki.moveForward(amount);
        }
        else
        {
          sparki.moveForward();
        }
        break;
      case 'B':
        if (paramCount > 0)
        {
          int amount = paramArray[0];
          sparki.moveBackward(amount);
        }
        else
        {
          sparki.moveBackward();
        }
        break;
      case 'L':
        if (paramCount > 0)
        {
          int amount = paramArray[0];
          sparki.moveLeft(amount);
        }
        else
        {
          sparki.moveLeft();
        }
        break;
      case 'R':
        if (paramCount > 0)
        {
          int amount = paramArray[0];
          sparki.moveRight(amount);
        }
        else
        {
          sparki.moveRight();
        }
        break;
      case 'S': sparki.moveStop(); break;
      case 'C':
        if (paramCount > 0)
        {
          float amount = paramArray[0] / 10.0;
          sparki.gripperClose(amount);
        }
        else
        {
          sparki.gripperClose();
        }
        break;
      case 'O':
        if (paramCount > 0)
        {
          float amount = paramArray[0] / 10.0;
          sparki.gripperOpen(amount);
        }
        else
        {
          sparki.gripperOpen();
        }
        break;
      case 'P': sparki.gripperStop();  break;
      case 'G':
        if (paramCount > 0)
        {
          int sendCode = paramArray[0];
          if (sendCode < 256 && sendCode >= 0)
          {
            sparki.sendIR(sendCode);
          }
        }
        else
        {
          sparki.sendIR(130);
        }
        break;
      case 'H':
        if (paramCount == 0)
        {
          sparki.beep();
        }
        else if (paramCount == 1)
        {
          sparki.beep(paramArray[0]);
        }
        else
        {
          sparki.beep(paramArray[0],param1multiple);
        }
        break;
      case 'V':
        if (paramCount == 0 && paramArray[0] > 0)
        {
          sparki.speed(paramArray[0]);
        }
        else
        {
          sparki.speed(20);
        }
      case '+':
        if (paramCount == 0)
        {
          currentLightColor++;
        }
        if (paramCount == 1)
        {
          currentLightColor = paramArray[0];
        }
        if (paramCount < 2)
        {
          if (currentLightColor > 3) currentLightColor = 0;
          switch (currentLightColor)
          {
            case 0: sparki.RGB(RGB_OFF); break;
            case 1: sparki.RGB(RGB_RED); break;
            case 2: sparki.RGB(RGB_WHITE); break;
            case 3: sparki.RGB(RGB_BLUE); break;
          }
        }
        else if (paramCount == 3)
        {
          sparki.RGB(paramArray[0],paramArray[1],paramArray[2]);
        }
        break;
      case '\\':
        if (paramCount == 0)
        {
          servoPosition = servoPosition - 5;
        }
        else
        {
          servoPosition = servoPosition - paramArray[0];
        }
        if (servoPosition < SERVO_LEFT) servoPosition = SERVO_LEFT;
        sparki.servo(servoPosition);
        break;
      case '/':
        if (paramCount == 0)
        {
          servoPosition = servoPosition + 5;
        }
        else
        {
          servoPosition = servoPosition + paramArray[0];
        }
        if (servoPosition > SERVO_RIGHT) servoPosition = SERVO_RIGHT;
        sparki.servo(servoPosition);
        break;
      case '|':
        if (paramCount == 0)
        {
          servoPosition = SERVO_CENTER;
        }
        else
        {
          servoPosition = paramArray[0];
          if (servoPosition > SERVO_RIGHT) servoPosition = SERVO_RIGHT;
          else if (servoPosition < SERVO_LEFT) servoPosition = SERVO_LEFT;
        }
        sparki.servo(servoPosition);
        break;
      case 'l':
      case 'b':
        /* stacking codes with conditional breaks is not my favorite way to handle 'case'
         * selection but it saves a little bit of space */
        sparki.motorRotate(MOTOR_LEFT,-2*paramArray[0]+1,param1multiple,paramArray[2]);
        if (serialCode == 'l')
          break;
      case 'r':
        sparki.motorRotate(MOTOR_RIGHT,paramArray[0]*2-1,param1multiple,paramArray[2]);
        break;
    }
  }
  
  if (newData)
  {
    sparki.clearLCD();
    
    sparki.print("IR code: ");
    sparki.println(IRCode);
    sparki.print("Hits: ");
    sparki.println(hits);
    sparki.print("Serial code: ");
    sparki.println(serialCode);
    
    sparki.print("Params: ");
    if (serialCode != '0')
    {
      for (char i = 0; i < MAX_BUFFER; i++)
      {
        char m = additionalParams[i];
        if (m > 0)
        {
          sparki.print(m);
        }
        else
        {
          sparki.print('_');
        }
      }
    }
    sparki.println();
    sparki.print("Parameter count: ");
    sparki.println(paramCount);
    
    sparki.updateLCD();
  }
  
  char tBuffer[6];
  Serial1.write('<');
  
  float floatVals[] = {sparki.accelX(), sparki.accelY(), sparki.accelZ(), sparki.magX(), sparki.magY(), sparki.magZ(), sparki.compass()};
  for (char j = 0; j < 7; j++)
  {
    Serial1.write(dtostrf(floatVals[j],5,2,tBuffer));
    writeSeparator(); // saves space over individual "Serial1.write('_');" commands
  }

  int intVals[] = {sparki.ping(), sparki.edgeLeft(), sparki.lineLeft(), sparki.lineCenter(), sparki.lineRight(), sparki.edgeRight(), sparki.lightLeft(), sparki.lightCenter(), sparki.lightRight()};
  for (char j = 0; j < 9; j++)
  {
    String intVal = String(intVals[j]);
    intVal.toCharArray(tBuffer,20);
    Serial1.write(tBuffer);
    writeSeparator(); // saves space over individual "Serial1.write('_');" commands
  }

  Serial1.write(">\n");
  newData = false;
  if (newData)
  {
    sparki.clearLCD();
    
    sparki.print("IR code: ");
    sparki.println(IRCode);
    sparki.print("Hits: ");
    sparki.println(hits);
    sparki.print("Serial code: ");
    sparki.println(serialCode);
    
    sparki.print("Params: ");
    if (serialCode != '0')
    {
      for (char i = 0; i < MAX_BUFFER; i++)
      {
        char m = additionalParams[i];
        if (m > 0)
        {
          sparki.print(m);
        }
        else
        {
          sparki.print('_');
        }
      }
    }
    sparki.println();
    
    sparki.updateLCD();
  }
}
